/* ------------------------------------------------
   Mandelbrot set
   z = z * z + c
   (C) 1992,94 by V.A.Voilevitch
---------------------------------------------------*/

#include "mandelbr.h"

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <graphics.h>

char *methods[] = { "OLD",
                    "MODERN",
                    "MIXED"
                  };


/*
 class MANDELBROT
 */

//-----------------------------------------------------------------
MANDELBROT::MANDELBROT(void) : rb(-2.5), re(0.5), ib(-1.25), ie(1.25),
                               sizx(100), sizy(100),
                               art(OLD), backgr(0),
                               it(100), st(0), ipc(1)
{
}


//-----------------------------------------------------------------
int MANDELBROT::range(void) const
{
  if( rb > re                    ||
      ib > ie                    ||
      sizx <= 0                  ||
      sizy <= 0                  ||
      sizx > 1 + getmaxx()       ||
      sizy > 1 + getmaxy()       ||
      it <= 0                    ||
      st < 0                     ||
      st >= it                   ||
      ipc < 1    && art==MIXED   ||
      backgr < 0                 ||
      backgr > getmaxcolor()
    )
    return No;
  else
    return Yes;
}

//-----------------------------------------------------------------
int MANDELBROT::show(void) const
{
  if( !range() )  return No;

  setfillstyle(SOLID_FILL, backgr);
  bar(0, 0, sizx-1, sizy-1);

  int rest,                  // remain iterations
      dit;                   // iterations per color by OLD method
  int maxc = getmaxcolor();  // max color
  int numc = 1 + maxc;       // color number

  double si = (ie - ib) / sizy;    // steps
  double sr = (re - rb) / sizx;
  double ci, cr;                   // current point
  double zr,zi;                    // current z
  double iq,rq;                    // temporary

  register setcol,                      // evaluated color
           lt,                          // current iteration
           dt;                          // current iteration - it0

  switch(art)
  { case OLD   : dit = ((it%maxc)?1:0) + it / maxc;
                 rest = maxc - it % numc;
                 break;
    case MODERN: rest = maxc - it % numc;
                 break;
    case MIXED : int ai = it / ipc;
                 rest = maxc - ai % numc;
                 break;
  }


  ci = ib - si;
  for(int j=0;j<sizy;j++)
  { ci += si;
    cr = rb - sr;
    for(int i=0;i<sizx;i++)
    { cr += sr;
      zi = ci;
      zr = cr;
      iq = zi * zi;
      rq = zr * zr;
      lt = 0;
      do
      { zi = 2 * zr * zi + ci;
        zr = rq - iq + cr;
        iq = zi * zi;
        rq = zr * zr;
        lt++;
      } while( lt < it && iq + rq <= 4 );
      if     ( lt < st ) setcol = backgr;
      else if( art==OLD )
	setcol = ( lt + rest) / dit;
      else if( art==MODERN )
	setcol = (lt + rest) % numc;
      else if( art==MIXED )
	setcol = (lt / ipc + rest) % numc;
      if( setcol != backgr )
        putpixel(i,j,setcol);
    }
  }
  putchar('\a');
  return Ok;
}

//-----------------------------------------------------------------
int MANDELBROT::init(void)
{ double  f_dbl;
  puts("\n Area of picture...");
  printf("start real : ");
  scanf("%lf",&f_dbl);
  rb = f_dbl;
  printf("end real   : ");
  scanf("%lf",&f_dbl);
  re = f_dbl;
  printf("start image: ");
  scanf("%lf",&f_dbl);
  ib = f_dbl;
  printf("end image  : ");
  scanf("%lf",&f_dbl);
  ie = f_dbl;

  puts("\n Size of picture...");
  printf("x size (1..%d) : ",1+getmaxx());
  scanf("%d",&sizx);
  printf("y size (1..%d) : ",1+getmaxy());
  scanf("%d",&sizy);

  puts( "\n Method of painting :\n"
        "\t 0 - old\n"
        "\t 1 - modern\n"
        "\t 2 - mixed\n"
      );
  char cm;
  do cm=getch(); while(cm<'0' || cm>'2');
  switch(cm)
  { case '0': art = OLD;
              break;
    case '1': art = MODERN;
              break;
    case '2': art = MIXED;
              break;
    default:  art = MODERN;
  }
  printf("%s\n",methods[art]);
  printf(   "background color (0..%d) : ",getmaxcolor());
  scanf("%d",&backgr);
  printf(   "number of iterations     : ");
  scanf("%d",&it);
  printf(   "start iteration (0..%d)  : ",it-1);
  scanf("%d",&st);
  if(art==MIXED)
  { printf( "iterations per color (1....)    : ");
    scanf("%d",&ipc);
  }

  return Yes;
}


//-----------------------------------------------------------------
int MANDELBROT::init(char *filename)
{ char fname[300];
  if(!strcmp(filename,"?"))
  { printf("Enter file name : ");
    fgets(fname, 299, stdin);
    fname[strlen(fname)-1] = '\0';
  }
  else
  { strncpy(fname,filename,299);
  }
  double  f_dbl1, f_dbl2;
  FILE * f = fopen(fname,"rt");
  if(!f)
  { printf( "Cannot open for reading file %s :\n"
            "-> %s\n"
            "...type any key to continue...\n",
            fname, sys_errlist[errno]
          );
    getch();
    return No;
  }
  fscanf(f,"%lf %lf",&f_dbl1,&f_dbl2);
  rb = f_dbl1;
  re = f_dbl2;
  fscanf(f,"%lf %lf",&f_dbl1,&f_dbl2);
  ib = f_dbl1;
  ie = f_dbl2;
  fscanf(f,"%d %d",&sizx,&sizy);
  int cc;
  fscanf(f,"%d",&cc);
  switch(cc)
  { case 0:  art = OLD;
             break;
    case 1:  art = MODERN;
             break;
    case 2:  art = MIXED;
             break;
    default: art = MODERN;
  }
  fscanf(f,"%d",&backgr);
  fscanf(f,"%d",&it);
  fscanf(f,"%d",&st);
  if(art==MIXED)
    fscanf(f,"%d",&ipc);

  fclose(f);
  return Yes;
}


//-----------------------------------------------------------------
int MANDELBROT::prompt(void)
{
  puts("----------------------\n"
       "You entered:");
  printf( "real  : %20.15lf ... %20.15lf\n"
          "image : %20.15lf ... %20.15lf\n"
	  "%d iterations (from %d)\n"
          "background = %d\n",
	  rb,re,
          ib,ie,
          it,st,
          backgr
        );
  if(art==MIXED)
     printf("%d iterations per color\n",ipc);
  printf("Picture %d x %d\n"
         "method %s",
         sizx, sizy,
         methods[art]
        );

  puts("\n  Type any key or ESC to quit...");
  return getch()!=27;
}

//-----------------------------------------------------------------
int MANDELBROT::save(char *filename)
{ if(!filename)
  { char fname[300];
    printf("Enter file name : ");
    fflush(stdin);
    fgets(fname, 299, stdin);
    fname[strlen(fname)-1] = '\0';
    filename = fname;
  }
  FILE * f = fopen(filename,"wt");
  if(!f)
  { printf( "Cannot open for writing file %s :\n"
            "-> %s\n"
            "...type any key to continue...\n",
            filename, sys_errlist[errno]
          );
    getch();
    return No;
  }
  fprintf(f,"%lf %lf\n"
            "%lf %lf\n"
            "%d %d\n"
            "%d\n"
            "%d\n"
            "%d\n"
            "%d\n",
            rb,re,
            ib,ie,
            sizx,sizy,
            int(art),
            backgr,
            it,
            st
         );
  if(art==MIXED)
    fprintf(f,"%d\n",ipc);
  fclose(f);
  return Yes;
}
